
// Copyright (C) Tom Horsley, 1997. All rights reserved.

#include <windows.h>
#include "base-exception.h"
#include "ftrace.h"

BaseException::BaseException() throw()
{
   return;
}

BaseException::BaseException(const BaseException& rhs) throw()
: exception(rhs)
{
   return;
}

BaseException&
BaseException::operator=(const BaseException& rhs) throw()
{
   exception::operator=(rhs);
   description = rhs.description;
   return *this;
}

BaseException::~BaseException() throw()
{
   return;
}

void
BaseException::Printf(const char * fmt, ...) {
   va_list vl;
   va_start(vl, fmt);
   VPrintf(fmt, vl);
   va_end(vl);
}

void
BaseException::VPrintf(const char * fmt, va_list vl) {
   char buf[1024];
   wvsprintf(buf, fmt, vl);
   description.append(&buf[0]);
}

void
BaseException::AddCalledFromInfo(const FunctionTracer & tm) {
#if ACTIVATE_FTRACE >= 1
   ConditionalNewline();
   description.append("Called from ");
   description.append(tm.GetFunctionTraceString());
#endif
}

void
BaseException::AddThrowLocation(const char * file, int line) {
   ConditionalNewline();
   Printf("Thrown from \"%s\":%d", file, line);
}

const std::string &
BaseException::GetDescription() {
   return description;
}

const char *
BaseException::what() const throw() {
   return description.c_str();
}

void
BaseException::ConditionalNewline() {
   if ((description.size() > 0) &&
       (description[description.size() - 1] != '\n')) {
      description.append(1, '\n');
   }
}
