
// Copyright (C) Tom Horsley, 1997. All rights reserved.

#define WINVER 0x400

#include <windows.h>
#include <ras.h>
#include <exception>
#include "ntp-service.h"
#include "wrap-winsock.h"
#include "enable-set-time.h"
#include "raise-priority.h"
#include "ntp-packet.h"
#include "log-directory.h"
#include "wrap-socket.h"
#include "regstuff.h"
#include "lasterr-exception.h"
#include "simple-exception.h"

typedef DWORD APIENTRY RasEnumConnections_func(LPRASCONNA, LPDWORD, LPDWORD);
typedef DWORD APIENTRY RasGetConnectStatus_func(HRASCONN, LPRASCONNSTATUSA);
typedef DWORD APIENTRY RasGetErrorString_func(UINT, LPSTR, DWORD);

bool
NtpService::IsRasActive()
{
   // Look at all the trouble we have to go to to find out if there is an
   // active RAS connection while avoiding autodial (There is actually a
   // wonderful service that will allow me to wait on an event, but it only
   // works under NT 4.0 or greater, sigh...)

   // Now we go to even *more* trouble to avoid having to have RAS installed
   // in case this routine is never called (A lot of folks have direct
   // network connections and don't want to go to the trouble of installing
   // RAS on their workstations). The first time we call this, try to
   // LoadLibrary the rasapi32 lib and dynamically lookup the RAS functions
   // we need to call...

   if (! ras_dll_loaded) {
      ras_dll_loaded = true;
      ras_dll_handle = LoadLibrary("rasapi32");
      if (ras_dll_handle == NULL) {
         GetLog()->LogString("Unable to LoadLibrary the rasapi32 DLL");
         return false;
      }
      ras_enum_connections_ptr =
         GetProcAddress(ras_dll_handle, "RasEnumConnectionsA");
      ras_get_connection_status_ptr =
         GetProcAddress(ras_dll_handle, "RasGetConnectStatusA");
      ras_get_error_string_ptr =
         GetProcAddress(ras_dll_handle, "RasGetErrorStringA");
   }
   if ((ras_dll_handle == NULL) || (ras_enum_connections_ptr == NULL) ||
       (ras_get_connection_status_ptr == NULL)) {
      return false;
   }
   RasEnumConnections_func * RasEnumConnections =
      (RasEnumConnections_func *)ras_enum_connections_ptr;
   RasGetConnectStatus_func * RasGetConnectStatus =
      (RasGetConnectStatus_func *)ras_get_connection_status_ptr;
   RasGetErrorString_func * RasGetErrorString =
      (RasGetErrorString_func *)ras_get_error_string_ptr;
   RASCONN * connection_poop;
   RASCONN dummy;
   DWORD connection_count;
   DWORD connection_size = 0;
   memset((void *)&dummy, 0, sizeof(dummy));
   dummy.dwSize = sizeof(RASCONN);
   RasEnumConnections(&dummy, &connection_size, &connection_count);
   if (connection_size < sizeof(RASCONN)) {
      if (connection_size == 0) {
         return false;
      }
      connection_size = sizeof(RASCONN);
   }
   connection_count = (connection_size + sizeof(RASCONN) - 1) /
                         sizeof(RASCONN);
   connection_poop = new RASCONN [connection_count];
   connection_size = connection_count * sizeof(RASCONN);
   memset((void *)connection_poop, 0, connection_size);
   connection_poop[0].dwSize = sizeof(RASCONN);
   DWORD errcode = RasEnumConnections(connection_poop, &connection_size,
                                      &connection_count);
   if (errcode != 0) {
      static char errmsg[512];
      errmsg[0] = '\0';
      if ((ras_get_error_string_ptr != NULL) &&
          (RasGetErrorString(errcode, errmsg, sizeof(errmsg)) == 0)) {
         throw new SimpleException(errmsg);
      } else {
         throw new SimpleException("RasEnumConnections error");
      }
   }
   bool found_valid_ras = false;
   for (DWORD ri = 0; ri < connection_count; ++ri) {
      RASCONNSTATUS ras_stat;
      memset((void *)&ras_stat, 0, sizeof(ras_stat));
      ras_stat.dwSize = sizeof(ras_stat);
      if (RasGetConnectStatus(connection_poop[ri].hrasconn, &ras_stat) == 0) {
         if (ras_stat.rasconnstate == RASCS_Connected) {
            found_valid_ras = true;
            break;
         }
      }
   }
   delete [] connection_poop;
   return found_valid_ras;
}
