#ifndef log_directory_h
#define log_directory_h

// Copyright (C) Tom Horsley, 1997. All rights reserved.

#include <stdio.h>
#include <windows.h>
#include <exception>
#include "wrap-critical.h"
#include "log-stuff.h"

// A class to manage logging entries to a family of log files in a
// given directory. (You get a new log files every day).
//
class LogDirectory : public LogStuff {
public:

      // Constructor takes directory name and suffix string for log
      // files. Also takes a max count for number of log files to keep.
      //
   LogDirectory(
      const char * init_dirname,
      const char * init_suffix,
      DWORD init_log_file_count);

   ~LogDirectory();

   virtual void
   LogException(std::exception & ee);

   virtual void
   LogString(const char * str);

   virtual void
   LogPrintf(const char * fmt, ...);

   virtual bool
   ValidLog();

private:

   void
   InternLogString(const char * str);

   void
   Lock();

   void
   UnLock();

   DeclareCriticalSection log_lock;

   void
   DeleteOldLogFiles();

   char * curlog_name;

   FILE * curlog;

   const char * dirname;
   const char * suffix;

   DWORD log_file_count;

   SYSTEMTIME now;
};

#endif // log_directory_h
