#ifndef ntp_events_h
#define ntp_events_h

#include <windows.h>

// This base class defines the interface used to diddle the events that
// the ntpoke.exe and ntptray.exe programs use to interact with NTPTime.
//
// This base class merely provides a dummy interface that does nothing.
// If the AllowPoke registry key is 0, then this dummy class is used
// everywhere.
//
// You can think of the state events in here as kind of "idiot lights" on
// the NTPTime dashboard. Only one event is ever in a set state at any time,
// and the tray icon app simply has to wait on all the other events to
// detect a state change and tell the user about it.
//
class NtpEvents {
public:

   NtpEvents();

   virtual
   ~NtpEvents();

      // Check this after the constructor to see if everything went OK
      //
   virtual bool
   IsValid();

      // If the poke event exists, append it to the ev array, returning the
      // new count (same as input count if no poke event). If a stop event
      // exists, append it following the poke event. This base class
      // implementation always returns input evcount value and leaves ev
      // unmodified.
      //
   virtual DWORD
   AppendExtraEvents(HANDLE ev[], DWORD evcount);

      // Reset the poke event.
      //
   virtual void
   ResetPoke();

      // Set the "init state" event, reset any other set state.
      //
   virtual void
   SetInitState();

      // Set the "sync OK" event, reset any other set state.
      //
   virtual void
   SetSyncOK();

      // Set the "sync warning" event, reset any other set state.
      //
   virtual void
   SetSyncWarn();

      // Set the "shutting down" event, reset any other set state.
      //
   virtual void
   SetShutdownState();
};

#endif // ntp_events_h
