
// Copyright (C) Tom Horsley, 1997. All rights reserved.

#include <string.h>
#include <ctype.h>
#include "ntp-service.h"
#include "lasterr-exception.h"
#include "log-directory.h"

int WINAPI
WinMain(
   HINSTANCE hInstance,
   HINSTANCE hPrevInstance,
   LPSTR lpCmdLine,
   int nCmdShow)
{
   NtpService ns(hInstance);
   try {
      for ( ; ; ) {
         LPSTR cmdarg = NULL;
         if (lpCmdLine != NULL) {
            while (isspace(*lpCmdLine)) ++lpCmdLine;
            cmdarg = lpCmdLine;
            while ((*lpCmdLine != '\0') && (! isspace(*lpCmdLine))) ++lpCmdLine;
            if (*lpCmdLine != '\0') {
               *lpCmdLine++ = '\0';
            }
         }
         if ((cmdarg != NULL) && (cmdarg[0] != '\0')) {
            char * msgbuf = new char [4096];
            if (cmdarg[0] == '/') {
               cmdarg[0] = '-';
            }
            if (cmdarg[1] == '/') {
               cmdarg[1] = '-';
            }
            if ((strcmp(cmdarg, "--help") == 0) ||
                (strcmp(cmdarg, "-h") == 0) ||
                (strcmp(cmdarg, "-H") == 0) ||
                (strcmp(cmdarg, "-?") == 0)) {
               strcpy(msgbuf,
"ntptime - win32 simple NTP time setting service\n\
usage: ntptime [--help] [[--noras] --install[q]] [--uninstall[q]]\n\
\n\
--help          print this message and exit\n\
--noras         do not install dependency on the RAS service\n\
--install       install this as a service on windows NT\n\
--installq      just like install, but does not popup a message on success\n\
--uninstall     remove this as a windows NT service\n\
--uninstallq    just like uninstall, but does not popup a message on success\n\
--version       print version and copyright information\n\
--w95           debug option to run as a Windows 95 app even on NT\n\
\n\
ntptime runs as a service on NT/2000/XP and a normal process on Windows\n\
95/98/ME.  It sets the system time periodically from a network NTP\n\
server. It can be configured to wait for a RAS connection before trying to\n\
set the time, set the time only once, or set the time periodically as long\n\
as the network connection is active.\n");
               MessageBox(NULL, msgbuf, "NTPTime", MB_OK);
               return 0;
            } else if (strcmp(cmdarg, "--w95") == 0) {
               ns.RunOnConsole();
            } else if ((strcmp(cmdarg, "--version") == 0) ||
                       (strcmp(cmdarg, "-v") == 0) ||
                       (strcmp(cmdarg, "-V") == 0)) {
               strcpy(msgbuf, "ntptime version ");
               strcat(msgbuf, NtpService::ntpversion);
               strcat(msgbuf, "\nCopyright (C) Tom Horsley 1997-2004.\n\
Unlimited distribution and use of unaltered program allowed.\n\
\n\
Tom.Horsley@worldnet.att.net\n");
               MessageBox(NULL, msgbuf, "NTPTime", MB_OK);
               return 0;
            } else if ((strcmp(cmdarg, "--noras") == 0) ||
                       (strcmp(cmdarg, "-r") == 0) ||
                       (strcmp(cmdarg, "-R") == 0)) {
               ns.SetUseRasDependency(false);
            } else if ((strncmp(cmdarg, "--install", 9) == 0) ||
                       (strcmp(cmdarg, "-i") == 0) ||
                       (strcmp(cmdarg, "-I") == 0)) {
               if (ns.IsRunningOnNT()) {
                  if (ns.Install(SERVICE_AUTO_START)) {
                     if (strcmp(cmdarg, "--installq") != 0) {
                        strcpy(msgbuf, "NTP service successfully installed\n");
                        MessageBox(NULL, msgbuf, "NTPTime", MB_OK);
                     }
                     return 0;
                  } else {
                     if (ns.IsInstalled()) {
                        strcpy(msgbuf,
                           "Unable to install service - already installed\n");
                     } else {
                        strcpy(msgbuf,
                           "Unable to install service - check privileges\n");
                     }
                     MessageBox(NULL, msgbuf, "NTPTime",
                                MB_OK | MB_ICONEXCLAMATION);
                     return 2;
                  }
               } else {
                  fputs("--install not required on Windows 95\n", stderr);
                  return 2;
               }
            } else if ((strncmp(cmdarg, "--uninstall", 11) == 0) ||
                       (strcmp(cmdarg, "-u") == 0) ||
                       (strcmp(cmdarg, "-U") == 0)) {
               if (ns.IsRunningOnNT()) {
                  if (ns.Uninstall()) {
                     if (strcmp(cmdarg, "--uninstallq") != 0) {
                        strcpy(msgbuf, "NTP service successfully removed\n");
                        MessageBox(NULL, msgbuf, "NTPTime", MB_OK);
                     }
                     return 0;
                  } else {
                     if (! ns.IsInstalled()) {
                        strcpy(msgbuf,
                               "Unable to remove service - not installed\n");
                     } else {
                        strcpy(msgbuf,
                               "Unable to remove service - check privileges\n"
                               "(perhaps the service is currently running?)\n");
                     }
                     MessageBox(NULL, msgbuf, "NTPTime",
                                MB_OK | MB_ICONEXCLAMATION);
                     return 2;
                  }
               } else {
                  strcpy(msgbuf, "--uninstall not required on Windows 95\n");
                  MessageBox(NULL, msgbuf, "NTPTime",
                             MB_OK | MB_ICONEXCLAMATION);
                  return 2;
               }
            } else {
               strcpy(msgbuf, "Unrecognized argument: ");
               strcat(msgbuf, cmdarg);
               MessageBox(NULL, msgbuf, "NTPTime",
                          MB_OK | MB_ICONEXCLAMATION);
               return 2;
            }
         } else {
            if (! ns.StartService()) {
               return 2;
            }
            return 0;
         }
      }
   } catch (std::exception * ee) {
      ns.GetLog()->LogException(*ee);
      delete ee;
      return 2;
   }
   return 0;
}
