
// Copyright (C) Tom Horsley, 1997. All rights reserved.

#include "regstuff.h"
#include "lasterr-exception.h"
#include "simple-exception.h"

RegistryKeyBase::RegistryKeyBase(HKEY parent_key, LPCTSTR key_name)
{
   CommonInit(parent_key, key_name);
}

RegistryKeyBase::RegistryKeyBase(RegistryKeyBase & parent_key, LPCTSTR key_name)
{
   CommonInit(parent_key.my_key, key_name);
}

RegistryKeyBase::~RegistryKeyBase()
{
   return;
}

DWORD
RegistryKeyBase::GetRegDwordValue(const char * valname)
{
   DWORD rval;
   DWORD fmt;
   DWORD len = sizeof(rval);
   if (RegQueryValueEx(my_key, valname, 0, &fmt, (LPBYTE)&rval, &len) !=
           ERROR_SUCCESS) {
      throw new SimpleException("missing registry entry");
   } else if (fmt != REG_DWORD) {
      throw new SimpleException("unexpected format registry value");
   }
   return rval;
}

char *
RegistryKeyBase::GetRegSzValue(const char * valname)
{
   char * rval;
   DWORD fmt;
   DWORD len = 0;
   if (RegQueryValueEx(my_key, valname, 0, &fmt, NULL, &len) !=
           ERROR_SUCCESS) {
      throw new SimpleException("missing registry entry");
   } else if (fmt != REG_SZ) {
      throw new SimpleException("unexpected format registry value");
   }
   rval = new char [len];
   if (RegQueryValueEx(my_key, valname, 0, &fmt, (LPBYTE)rval, &len) !=
           ERROR_SUCCESS) {
      delete [] rval;
      throw new SimpleException("missing registry entry");
   } else if (fmt != REG_SZ) {
      delete [] rval;
      throw new SimpleException("unexpected format registry value");
   }
   return rval;
}

char *
RegistryKeyBase::GetRegExpandSzValue(const char * valname) {
   char * rval;
   DWORD fmt;
   DWORD len = 0;
   if (RegQueryValueEx(my_key, valname, 0, &fmt, NULL, &len) !=
           ERROR_SUCCESS) {
      throw new SimpleException("missing registry entry");
   } else if (! ((fmt == REG_SZ) || (fmt == REG_EXPAND_SZ))) {
      throw new SimpleException("unexpected format registry value");
   }
   rval = new char [len];
   if (RegQueryValueEx(my_key, valname, 0, &fmt, (LPBYTE)rval, &len) !=
           ERROR_SUCCESS) {
      delete [] rval;
      throw new SimpleException("missing registry entry");
   } else if (! ((fmt == REG_SZ) || (fmt == REG_EXPAND_SZ))) {
      delete [] rval;
      throw new SimpleException("unexpected format registry value");
   }
   if (fmt == REG_EXPAND_SZ) {
      char dummy[1];
      char * origrval = rval;
      DWORD newlen = ExpandEnvironmentStrings(rval, dummy, 1);
      if (newlen > 0) {
         char * newrval = new char [newlen + 1];
         DWORD finlen = ExpandEnvironmentStrings(rval, newrval, newlen);
         if (newlen >= finlen) {
            delete [] rval;
            rval = newrval;
         }
      }
      if (origrval == rval) {
         throw new SimpleException(
            "unable to expand environment variables in registry");
      }
   }
   return rval;
}

bool
RegistryKeyBase::QueryValue(
   const char * valname,
   DWORD & val_format,
   DWORD & val_length)
{
   val_length = 0;
   val_format = 0;
   return (
      RegQueryValueEx(my_key, valname, 0, &val_format, NULL, &val_length) ==
         ERROR_SUCCESS
          );
}

RegistryKeyBase::RegistryKeyBase(HKEY init_key)
{
   my_key = init_key;
}

void
RegistryKeyBase::CommonInit(HKEY parent_key, LPCTSTR key_name)
{
   if (RegOpenKeyEx(parent_key, key_name, 0, KEY_READ, &my_key) !=
          ERROR_SUCCESS) {
      throw new SimpleException("missing registry key");
   }
}

RegistryPredefinedKey::RegistryPredefinedKey(HKEY init_key)
: RegistryKeyBase(init_key)
{
   return;
}

RegistryPredefinedKey::~RegistryPredefinedKey()
{
   Cleanup();
}

void
RegistryPredefinedKey::Cleanup()
{
   return;
}

RegistryClassesRoot::RegistryClassesRoot()
: RegistryPredefinedKey(HKEY_CLASSES_ROOT)
{
   return;
}

RegistryClassesRoot::~RegistryClassesRoot()
{
   return;
}

RegistryCurrentUser::RegistryCurrentUser()
: RegistryPredefinedKey(HKEY_CURRENT_USER)
{
   return;
}

RegistryCurrentUser::~RegistryCurrentUser()
{
   return;
}

RegistryLocalMachine::RegistryLocalMachine()
: RegistryPredefinedKey(HKEY_LOCAL_MACHINE)
{
   return;
}

RegistryLocalMachine::~RegistryLocalMachine()
{
   return;
}

RegistryUsers::RegistryUsers()
: RegistryPredefinedKey(HKEY_USERS)
{
   return;
}

RegistryUsers::~RegistryUsers()
{
   return;
}

RegistryKey::RegistryKey(HKEY parent_key, LPCTSTR key_name)
: RegistryKeyBase(parent_key, key_name)
{
   return;
}

RegistryKey::RegistryKey(RegistryKeyBase & parent_key, LPCTSTR key_name)
: RegistryKeyBase(parent_key, key_name)
{
   return;
}

RegistryKey::~RegistryKey()
{
   Cleanup();
}

void
RegistryKey::Cleanup()
{
   RegCloseKey(my_key);
}
