
// Copyright (C) Tom Horsley, 1997. All rights reserved.

#include <winsock2.h>
#include <windows.h>
#include "lasterr-exception.h"

LastErrorException::LastErrorException(
   DWORD init_error_number,
   const char * fmt,
   ...) throw() {
   va_list vl;
   va_start(vl, fmt);
   VPrintf(fmt, vl);
   va_end(vl);
   CommonFormat(init_error_number);
}

LastErrorException::LastErrorException(
   const char * fmt,
   ...) throw() {
   DWORD init_error_number = GetLastError();
   va_list vl;
   va_start(vl, fmt);
   VPrintf(fmt, vl);
   va_end(vl);
   CommonFormat(init_error_number);
}

LastErrorException::LastErrorException(const LastErrorException& rhs) throw()
: BaseException(rhs)
{
   return;
}

LastErrorException::~LastErrorException() throw()
{
   return;
}

LastErrorException &
LastErrorException::operator=(const LastErrorException & rhs) throw()
{
   BaseException::operator=(rhs);
   return *this;
}

// Sigh... Why Microsoft goes to all the trouble to define FormatMessage and
// then leaves out all the winsock messages is beyond me...
//
// This table allows me to lookup messages for winsock errors (this was
// taken from the list of errors actually generated by winsock services as
// given in the comments from the SDK header files on networking errors).
//
const LastErrorException::extra_errmsg_info
LastErrorException::winsock_errors[] = {
#include "wsaerr.h"
   {0, NULL}
};

void
LastErrorException::CommonFormat(DWORD errnum) throw() {
   char msgbuf[1024];
   const char * msgptr = &msgbuf[0];
   DWORD size = FormatMessage(
                   FORMAT_MESSAGE_FROM_SYSTEM,
                   NULL,
                   errnum,
                   MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),
                   msgbuf,
                   sizeof(msgbuf),
                   NULL);
   if (size == 0) {
      // Well, FormatMessage doesn't seem to know 'nuthin, lets see if
      // it is one of the messages in our table...
      for (int i = 0; winsock_errors[i].errmsg != NULL; ++i) {
         if (winsock_errors[i].errnum == errnum) {
            msgptr = winsock_errors[i].errmsg;
            size = strlen(msgptr);
         }
      }
   } else {
      // I want to put these messages inside other strings, but the
      // lovely message I actually get back has CRLF crap in it.
      // Change them all to spaces...
      for (int i = 0; i < size; ++i) {
         char c = msgbuf[i];
         if ((c == '\r') || (c == '\n')) {
            msgbuf[i] = ' ';
         }
      }
      while ((size > 0) && (msgbuf[size-1] == ' ')) {
         --size;
         msgbuf[size] = '\0';
      }
   }
   if (size == 0) {
      // Nobody knows 'nuthin about this error code, just report that...
      size = wsprintf(msgbuf, "Unknown error code 0x%08lx", errnum);
   }
   Printf(" : %s", msgptr);
}
