
// Copyright (C) Tom Horsley, 2004. All rights reserved.

// This is NTPTray.exe, the successor to ntpoke.exe. Aside from poking the
// server, NTPTray can act as a taskbar icon and monitor time sync status,
// examine log files, start and stop the service, etc.

// (With ntptray, all the junk that clutters up most programs is off in this
// separate program you don't have to run if you don't want to - optional
// code bloat! :-).

// PUT CODE HERE - should make the descriptions strings, about text,
// etc. into string resources...

// PUT CODE HERE - Add -w95 option to make ntptray act like it is running on
// windows 95 (don't use service manager, etc). This would match the --w95
// option on ntptime.

#define WINVER 0x400

#include <windows.h>
#include <shellapi.h>
#include <commctrl.h>
#include <string.h>
#include "resource.h"
#include "events.h"
#include "wrap-sc-handle.h"
#include "wrap-reg-key.h"
#include "wincmd.h"
#include "base-exception.h"
#include "simple-exception.h"
#include "lasterr-exception.h"

#define NTPTRAY_REGKEY "Software\\ntcli\\NTPTray\\CurrentVersion"

static bool show_help = false;

static long int poll_seconds = 15;

static char * log_viewer = "notepad.exe";
static bool delete_log_viewer = false;

static bool just_do_poke = false;

static HINSTANCE app_instance;

static char * AppName = "NTPTray";

static HWND main_window;

static HMENU main_menu;

static HICON app_icon;

static NOTIFYICONDATA nid;

static HANDLE poke_event = NULL;

#define MY_ICON_MSG (WM_APP+100)

static LRESULT CALLBACK WndProc(HWND, UINT, WPARAM, LPARAM);

static DWORD default_tmout = 15000;
static DWORD cur_tmout = 15000;

// Array of these tells us how to find state and set tray icon.
//
struct ntp_state_info {
      // The icon resource number for the tray icon that represents this
      // state.
      //
   int    icon_resource;

      // Description of this state for the tooltip message
      //
   char * desc;

      // The name of the event associated with this state.
      //
   char * event_name;

      // Icon used to represent this state in taskbar
      //
   HICON  hi;

      // When set, this event indicates we should be in this state.
      //
   HANDLE ev;

      // Track status of this event
      //
   bool ev_set;

      // Remember if we should popup a message when entering this state.
      //
   bool cur_pop_msg;
   bool def_pop_msg;

      // Resource number of the checkbox control in options window
      // for requesting message when switching to this state.
      //
   int ckres;
};

static struct ntp_state_info states[] = {
#define STATE_NONE 0
   {IDI_NONE_ICON, "NTPTime not running", EV_NTPTIME_CLOSE, NULL, NULL,
    false, false, false, IDC_NONE_CHECK},

#define STATE_WARN 1
   {IDI_WARN_ICON, "Problem setting time", EV_NTPTIME_WARN, NULL, NULL,
    false, false, false, IDC_WARN_CHECK},

#define STATE_INIT 2
   {IDI_INIT_ICON, "NTPTime initializing", EV_NTPTIME_INIT, NULL, NULL,
    false, false, false, IDC_INIT_CHECK},

#define STATE_OK 3
   {IDI_OK_ICON, "Time in sync", EV_NTPTIME_OK, NULL, NULL,
    false, false, false, IDC_OK_CHECK}
};

#define STATE_COUNT (sizeof(states)/sizeof(states[0]))

static char about_text[] = "\
NTPTray Version 4.3\r\n\
A taskbar application for interaction with\r\n\
NTPTime - a simple time setting client\r\n\
using the NTP internet protocol.\r\n\
\r\n\
The NTPTime programs are freeware,\r\n\
copyright Tom Horsley 1997-2004.";

static int last_state = -1;

static void CloseNtpEvents();

static HANDLE OpenGlobalOrNotEvent(char * local_name, DWORD af);

static void CheckNtpEvents();

static BOOL CALLBACK
AboutDlgProc(HWND hDlg, UINT iMsg, WPARAM wParam, LPARAM lParam);

static BOOL CALLBACK
OptionsDlgProc(HWND hDlg, UINT iMsg, WPARAM wParam, LPARAM lParam);

static void SynchronizeNow();

static void StopNTPTimeService();

static void StartNTPTimeService();

static void StopNTPTimeWindow();

static void StartNTPTimeWindow();

static void StartNTPTimeControlPanel();

static bool ThisIsWindowsNT();

static void BrowseNtpLogs();

static void RunOtherProgram(char * progname, char * opt_arg, bool quote_arg);

static ArgDef TrayArgs [] = {
   {"Help", WINCMD_BOOL, "Show help window",
    NULL, NULL, NULL, REG_DWORD,
    (void *)&show_help, WINCMD_UNSPECIFIED},
   {"Poke", WINCMD_BOOL, "Ask NTPTime to set time now",
    NULL, NULL, NULL, REG_DWORD,
    (void *)&just_do_poke, WINCMD_UNSPECIFIED},
   {"PopShut", WINCMD_BOOL, "Popup message when NTPTime shuts down",
    HKEY_CURRENT_USER, NTPTRAY_REGKEY, "PopShut", REG_DWORD,
    (void *)&states[STATE_NONE].def_pop_msg, WINCMD_UNSPECIFIED},
   {"PopWarn", WINCMD_BOOL, "Popup message when NTPTime has a problem",
    HKEY_CURRENT_USER, NTPTRAY_REGKEY, "PopWarn", REG_DWORD,
    (void *)&states[STATE_WARN].def_pop_msg, WINCMD_UNSPECIFIED},
   {"PopInit", WINCMD_BOOL, "Popup message when NTPTime initializing",
    HKEY_CURRENT_USER, NTPTRAY_REGKEY, "PopInit", REG_DWORD,
    (void *)&states[STATE_INIT].def_pop_msg, WINCMD_UNSPECIFIED},
   {"PopOk", WINCMD_BOOL, "Popup message when NTPTime sets time",
    HKEY_CURRENT_USER, NTPTRAY_REGKEY, "PopOk", REG_DWORD,
    (void *)&states[STATE_OK].def_pop_msg, WINCMD_UNSPECIFIED},
   {"PollSeconds", WINCMD_LONG, "Poll interval for NTPTime startup",
    HKEY_CURRENT_USER, NTPTRAY_REGKEY, "PollSeconds", REG_DWORD,
    (void *)&poll_seconds, WINCMD_UNSPECIFIED},
   {"LogViewer", WINCMD_STRING, "Program to use for viewing log files",
    HKEY_CURRENT_USER, NTPTRAY_REGKEY, "LogViewer", REG_SZ,
    (void *)&log_viewer, WINCMD_UNSPECIFIED},
};

#define TrayArgCount (sizeof(TrayArgs)/sizeof(TrayArgs[0]))

// ******************************************************************* WinMain

int WINAPI
WinMain(
   HINSTANCE hInstance,
   HINSTANCE hPrevInstance,
   PSTR szCmdLine,
   int iCmdShow) {

   Wincmd argv(szCmdLine);
   try {
      argv.InterpretArgs(TrayArgs, TrayArgCount);
   } catch ( BaseException * err ) {
      MessageBox(NULL, err->what(), AppName,
                 MB_OK | MB_ICONEXCLAMATION);
      delete err;
      return 2;
   }
   if (show_help) {
      std::string help_info("NTPTray app for interaction with NTPTime:\n\n");
      Wincmd::AppendHelpInfo(help_info, TrayArgs, TrayArgCount);
      MessageBox(NULL, help_info.c_str(), AppName,
                 MB_OK | MB_ICONINFORMATION);
      return 0;
   }
   if (just_do_poke) {
      try {
         SynchronizeNow();
      } catch ( BaseException * err ) {
         MessageBox(NULL, err->what(), AppName,
                    MB_OK | MB_ICONEXCLAMATION);
         delete err;
         return 2;
      }
      return 0;
   }

   // Run the program, displaying a taskbar icon, and waiting for the user
   // to tell us to exit via the popup menu.

   MSG msg;
   WNDCLASSEX wndclass;

   app_instance = hInstance;

   memset((void *)&wndclass, 0, sizeof(wndclass));

   app_icon = LoadIcon(hInstance, MAKEINTRESOURCE(IDI_APP_ICON));
   wndclass.cbSize        = sizeof (wndclass);
   wndclass.style         = CS_HREDRAW | CS_VREDRAW;
   wndclass.lpfnWndProc   = WndProc;
   wndclass.cbClsExtra    = 0;
   wndclass.cbWndExtra    = DLGWINDOWEXTRA;
   wndclass.hInstance     = hInstance;
   wndclass.hIcon         = app_icon;
   wndclass.hCursor       = LoadCursor(NULL, IDC_ARROW);
   wndclass.hbrBackground = (HBRUSH) (COLOR_3DFACE + 1);
   wndclass.lpszMenuName  = NULL;
   wndclass.lpszClassName = AppName;
   wndclass.hIconSm       = app_icon;

   RegisterClassEx(&wndclass);

   main_window = CreateWindow(AppName, "The Hidden NTPTray Window",
                              WS_OVERLAPPEDWINDOW, CW_USEDEFAULT,
                              CW_USEDEFAULT, CW_USEDEFAULT, CW_USEDEFAULT,
                              NULL, NULL, hInstance, NULL);

   ShowWindow(main_window, SW_HIDE);

   bool check_events = true;
   bool finished = false;
   DWORD tmout_count = 0;
   default_tmout = poll_seconds * 1000;
   cur_tmout = default_tmout;
   while (! finished) {
      if (check_events) {
         CheckNtpEvents();
         check_events = false;
      }
      while (PeekMessage(&msg, NULL, 0, 0, PM_REMOVE)) {
         if (msg.message == WM_QUIT) {
            finished = true;
            break;
         } else {
            TranslateMessage(&msg);
            DispatchMessage(&msg);
         }
      }
      if (! finished) {

         // Make arry holding just the events that are currently reset. If
         // the list is empty, make a timeout so we'll check events again
         // soon.

         HANDLE hands[STATE_COUNT];
         DWORD  cnt = 0;
         for (int i = 0; i < STATE_COUNT; ++i) {
            if (! states[i].ev_set) {
               hands[cnt++] = states[i].ev;
            }
         }

         DWORD tmout;
         if (cnt == 0) {
            tmout = cur_tmout;
            check_events = true;
         } else {
            tmout = INFINITE;
         }
         DWORD result = MsgWaitForMultipleObjects(cnt, hands, FALSE,
                                                  tmout, QS_ALLINPUT);

         // If we got a window event just process it. Any other return value
         // (timeout or event state change) we want to check the events
         // again.

         if (result != (WAIT_OBJECT_0 + cnt)) {
            check_events = true;
         }
         if (result == WAIT_TIMEOUT) {

            // If we timed out, check to see if we have been doing a lot of
            // that, if so increase the cur_tmout value if we had it set
            // small because we expected NTPTime to start soon - apparently
            // soon will never come.

            ++tmout_count;
            if (tmout_count > 30) {
               cur_tmout = default_tmout;
            }
         } else {
            // If we stopped waiting for any other reason, set timeout
            // count back to zero.
            tmout_count = 0;
         }
      }
   }

   Shell_NotifyIcon(NIM_DELETE, &nid);

   CloseNtpEvents();

   return msg.wParam;
}

// ******************************************************************* WndProc

LRESULT CALLBACK
WndProc(HWND hwnd, UINT iMsg, WPARAM wParam, LPARAM lParam) {
   POINT pt;

   try {
      switch(iMsg) {
      case WM_CREATE:
         main_menu = LoadMenu(app_instance, MAKEINTRESOURCE(IDR_MENU_POPUP));
         main_menu = GetSubMenu(main_menu, 0);
         SetMenuDefaultItem(main_menu, 0, TRUE);
         return 0;
      case WM_COMMAND:
         switch (LOWORD(wParam)) {
         case ID_NTPTIME_SYNCHRONIZENOW:

            // Poke the event that tells NTPTime to set time right now.

            SynchronizeNow();

            // After clicking this button, entering the "synchronized"
            // state won't be a surprise, so suppress that popup.

            states[STATE_OK].cur_pop_msg = false;
            break;
         case ID_NTPTIME_LOGFILES:

            // Popup browser on log directory with most recently modified log
            // file already selected.

            BrowseNtpLogs();
            break;
         case ID_NTPTIME_CONTROLPANEL:

            // spawn control.exe giving it ntptime.cpl as its argument on
            // command line.

            StartNTPTimeControlPanel();
            break;
         case ID_NTPTIME_OPTIONS:

            // Popup the "options" box.

            DialogBox(app_instance, MAKEINTRESOURCE(IDD_OPTION_DIALOG), hwnd,
                      OptionsDlgProc);
            break;
         case ID_NTPTIME_ABOUT:

            // Popup the "about" box.

            DialogBox(app_instance, MAKEINTRESOURCE(IDD_ABOUT_DIALOG), hwnd,
                      AboutDlgProc);
            break;
         case ID_NTPTIME_STARTNTPTIME:

            // If we are on NT try to start the NTPTime service, otherwise try
            // just starting ntptime.exe as a normal app.

            if (states[STATE_NONE].ev != NULL) {
               MessageBox(NULL, "NTPTime is already running.", AppName,
                          MB_OK | MB_ICONEXCLAMATION);
            } else {
               if (ThisIsWindowsNT()) {
                  StartNTPTimeService();
               } else {
                  StartNTPTimeWindow();
               }

               // If we just started NTPTime, monitor event creation
               // more frequently for a little while so we'll notice
               // when the events show up real soon now.

               cur_tmout = 500;

               // On the other hand, we shouldn't popup a message
               // about entring init state - it won't be unexpected
               // to change state when we just requested it.

               states[STATE_INIT].cur_pop_msg = false;
            }
            break;
         case ID_NTPTIME_STOPNTPTIME:

            // If we can find service running, try to stop it, otherwise see if
            // there is a top level window named NTPTime and send it a close
            // command.

            if (ThisIsWindowsNT()) {
               StopNTPTimeService();
            } else {
               StopNTPTimeWindow();
            }

            // When deliberately stopping the service, it won't be
            // a surprise to enter the shutdown state, so suppress
            // that popup.

            states[STATE_NONE].cur_pop_msg = false;
            break;
         case ID_NTPTIME_EXIT:
            PostQuitMessage(0);
            break;
         }
         return 0;
      case WM_DESTROY:
         PostQuitMessage(0);
         return 0;
      case MY_ICON_MSG:
         switch (lParam) {
         case WM_LBUTTONDBLCLK:
         case WM_LBUTTONDOWN:
            // Do time sync by default on single and double click.
            SynchronizeNow();
            states[STATE_OK].cur_pop_msg = false;
            break;
         case WM_RBUTTONDOWN:
            // Popup menu on right click.
            GetCursorPos(&pt);
            SetForegroundWindow(main_window);
            TrackPopupMenu(main_menu, 0, pt.x, pt.y, 0, main_window, NULL);
            PostMessage(main_window, WM_NULL, 0, 0);
            break;
         }
         return 0;
      }
   } catch ( BaseException * err ) {
      MessageBox(NULL, err->what(), AppName,
                 MB_OK | MB_ICONEXCLAMATION);
      delete err;
      return 0;
   }
   return DefWindowProc(hwnd, iMsg, wParam, lParam);
}

// ****************************************************** OpenGlobalOrNotEvent

static HANDLE
OpenGlobalOrNotEvent(char * local_name, DWORD af) {
   HANDLE rval = NULL;
   char globname[100];
   strcpy(globname, "Global\\");
   if ((strlen(globname) + strlen(local_name) + 1) < sizeof(globname)) {
      strcat(globname, local_name);
      rval = OpenEvent(af, FALSE, globname);
   }
   if (rval == NULL) {
      rval = OpenEvent(af, FALSE, local_name);
   }
   return rval;
}

// ************************************************************ CheckNtpEvents

static void
CheckNtpEvents() {

   DWORD notify_msg = NIM_MODIFY;
   int new_state = -1;

   // Go through states array. If we haven't yet built the icons, then
   // build them, if we don't have the events opened, try to open them.
   // Query each event to determine the state, and set the appropriate
   // icon in the system tray.

   for (int i = 0; i < STATE_COUNT; ++i) {
      if (states[i].hi == NULL) {
         states[i].hi = LoadIcon(app_instance,
                           MAKEINTRESOURCE(states[i].icon_resource));

         // If we haven't ever created icons before, we need to
         // add the icon in the tray, not simply modify it.

         notify_msg = NIM_ADD;
      }
      if (states[i].ev == NULL) {
         states[i].ev = OpenGlobalOrNotEvent(states[i].event_name, SYNCHRONIZE);
      }
      if (states[i].ev == NULL) {

         // If we don't have this event open, pretend it is set so we won't
         // try to wait for it in the event loop (where we wait on all the
         // events that are reset to see when one of them becomes set).

         states[i].ev_set = true;
      } else {

         // Wait on the event with zero timeout to see if it is currently
         // set.

         states[i].ev_set =
            (WaitForSingleObject(states[i].ev, 0) == WAIT_OBJECT_0);

         // If events exist, NTPTime must be running, so set timeout
         // back to default (we no longer need shorter polling interval).

         cur_tmout = default_tmout;
      }
      if ((new_state == -1) && states[i].ev_set) {
         new_state = i;
      }
   }
   if (new_state != last_state) {

      memset(&nid, 0, sizeof(nid));
      nid.cbSize = sizeof(NOTIFYICONDATA); 
      nid.hWnd = main_window;
      nid.hIcon = states[new_state].hi;
      nid.uFlags = NIF_TIP | NIF_ICON | NIF_MESSAGE;
      strncpy(nid.szTip, states[new_state].desc, sizeof(nid.szTip)); 
      nid.uCallbackMessage = MY_ICON_MSG;
      if (! Shell_NotifyIcon(notify_msg, &nid)) {
         MessageBox(NULL, "Cannot setup taskbar icon.", AppName,
                    MB_OK | MB_ICONEXCLAMATION);
         PostQuitMessage(0);
      }
      if (states[new_state].cur_pop_msg) {

         // This silly user not only wants his taskbar cluttered up, he also
         // wants a popup message when we go to this state, so pop up his
         // message.

         DWORD msgflag;
         if ((new_state == STATE_OK) || (new_state == STATE_INIT)) {
            msgflag = (MB_OK | MB_ICONINFORMATION);
         } else {
            msgflag = (MB_OK | MB_ICONEXCLAMATION);
         }
         MessageBox(NULL, states[new_state].desc, AppName, msgflag);
      }

      // Once state has actually changed, be sure to put cur_pop_msg flag
      // back to the same as def_pop_msg flag. Using these two sets of flags
      // allows us to suppress the popup when we know it is going to happen
      // next anyway and the popup would merely be redundant info.

      for (int i = 0; i < STATE_COUNT; ++i) {
         states[i].cur_pop_msg = states[i].def_pop_msg;
      }
      last_state = new_state;
   }
   if (states[STATE_NONE].ev_set && (states[STATE_NONE].ev != NULL)) {

      // Looks like the server is shutting down. It will be closing the
      // events it created, so I should close my handles to those events as
      // well.

      CloseNtpEvents();
   }
}

// ************************************************************ CloseNtpEvents

static void
CloseNtpEvents() {
   for (int i = 0; i < STATE_COUNT; ++i) {
      if (states[i].ev != NULL) {
         CloseHandle(states[i].ev);
         states[i].ev = NULL;
         states[i].ev_set = true;
      }
   }
   if (poke_event != NULL) {
      CloseHandle(poke_event);
      poke_event = NULL;
   }
}

// ************************************************************** AboutDlgProc

static BOOL CALLBACK
AboutDlgProc(HWND hDlg, UINT iMsg, WPARAM wParam, LPARAM lParam) {
   switch(iMsg) {
   case WM_INITDIALOG:

      // Get the about text in the edit control and the icon in
      // the button.

      SetWindowText(GetDlgItem(hDlg,IDC_ABOUT_EDIT), about_text);
      SendMessage(GetDlgItem(hDlg,IDC_ICON_BUTTON), BM_SETIMAGE,
                  (WPARAM)IMAGE_ICON,
                  (LPARAM)app_icon);
      return TRUE;
   case WM_COMMAND:
      switch(LOWORD(wParam)) {
      case IDOK:
      case IDCANCEL:
         EndDialog(hDlg, 0);
         return TRUE;
      case IDC_URL_BUTTON:
      case IDC_ICON_BUTTON:

         // Clicking the NTPTime icon or the url will launch your browser
         // and take you to the NTPTime home page (and pop down the
         // about box).

         ShellExecute(0, 0,
            "http://home.att.net/~Tom.Horsley/ntptime.html",
                      0, 0, SW_NORMAL);
         EndDialog(hDlg, 0);
         return TRUE;
      }
      break;
   }
   return FALSE;
}

// ************************************************************ OptionsDlgProc

static BOOL CALLBACK
OptionsDlgProc(HWND hDlg, UINT iMsg, WPARAM wParam, LPARAM lParam) {
   switch(iMsg) {
   case WM_INITDIALOG: {

         // Initialize widgets with values from option variables (which we
         // should have read from registry on startup).

         for (int i = 0; i < STATE_COUNT; ++i) {
            SendMessage(GetDlgItem(hDlg, states[i].ckres), BM_SETCHECK,
               (WPARAM)(states[i].def_pop_msg), 0);
         }
         SendMessage(GetDlgItem(hDlg, IDC_POLL_SPIN), UDM_SETPOS,
            0, (LPARAM)MAKELONG((short)poll_seconds, 0));
         SendMessage(GetDlgItem(hDlg, IDC_POLL_SPIN),
            UDM_SETRANGE, 0, (LPARAM) MAKELONG((short) UD_MAXVAL, (short) 1));
         SetWindowText(GetDlgItem(hDlg,IDC_VIEWLOG_EDIT), log_viewer);
      }
      return TRUE;
   case WM_COMMAND:
      switch(LOWORD(wParam)) {
      case IDOK: {

            // Get the settings out of the dialog box back into global
            // vars.

            int len = GetWindowTextLength(GetDlgItem(hDlg, IDC_POLL_EDIT));
            if (len > 0) {
               char * txt = new char [len + 1];
               GetWindowText(GetDlgItem(hDlg, IDC_POLL_EDIT), txt, len + 1);
               poll_seconds = atoi(txt);
               delete [] txt;
               if (poll_seconds <= 0) poll_seconds = 15;
               cur_tmout = poll_seconds * 1000;
               default_tmout = cur_tmout;
            }
            len = GetWindowTextLength(GetDlgItem(hDlg, IDC_VIEWLOG_EDIT));
            if (len > 0) {
               char * txt = new char [len + 1];
               GetWindowText(GetDlgItem(hDlg, IDC_VIEWLOG_EDIT), txt, len + 1);
               if (delete_log_viewer) delete [] log_viewer;
               log_viewer = txt;
               delete_log_viewer = true;
            }
            for (int i = 0; i < STATE_COUNT; ++i) {
               states[i].def_pop_msg =
                  (SendMessage(GetDlgItem(hDlg, states[i].ckres),
                               BM_GETCHECK, 0, 0) != 0);
               states[i].cur_pop_msg = states[i].def_pop_msg;
            }

            // copy options to current user's registry.

            HKEY tk;
            if (RegCreateKeyEx(HKEY_CURRENT_USER,
                               NTPTRAY_REGKEY,
                               0,
                               "",
                               REG_OPTION_NON_VOLATILE,
                               KEY_ALL_ACCESS,
                               NULL,
                               &tk,
                               NULL) == ERROR_SUCCESS) {
               WrapRegKey traykey(tk);
               DWORD temp;
               temp = states[STATE_NONE].def_pop_msg ? 1 : 0;
               RegSetValueEx(traykey, "PopShut", 0, REG_DWORD,
                             (CONST BYTE*)&temp, sizeof(temp));
               temp = states[STATE_WARN].def_pop_msg ? 1 : 0;
               RegSetValueEx(traykey, "PopWarn", 0, REG_DWORD,
                             (CONST BYTE*)&temp, sizeof(temp));
               temp = states[STATE_INIT].def_pop_msg ? 1 : 0;
               RegSetValueEx(traykey, "PopInit", 0, REG_DWORD,
                             (CONST BYTE*)&temp, sizeof(temp));
               temp = states[STATE_OK].def_pop_msg ? 1 : 0;
               RegSetValueEx(traykey, "PopOk", 0, REG_DWORD,
                             (CONST BYTE*)&temp, sizeof(temp));
               temp = (DWORD)poll_seconds;
               RegSetValueEx(traykey, "PollSeconds", 0, REG_DWORD,
                             (CONST BYTE*)&temp, sizeof(temp));
               RegSetValueEx(traykey, "LogViewer", 0, REG_SZ,
                             (CONST BYTE*)log_viewer,
                              strlen(log_viewer) + 1);
            }
         }
      case IDCANCEL:
         EndDialog(hDlg, 0);
         return TRUE;
      }
      break;
   }
   return FALSE;
}

// ************************************************************ SynchronizeNow

static void
SynchronizeNow() {
   if (poke_event == NULL) {
      poke_event = OpenGlobalOrNotEvent(EV_NTPTIME_POKE, EVENT_MODIFY_STATE);
   }
   if ((poke_event == NULL) || (! SetEvent(poke_event))) {
      throw new LastErrorException("Unable to make contact with NTPTime");
   }
}

// ******************************************************** StopNTPTimeService

static void
StopNTPTimeService() {

   // Presumably something in here will fail if we aren't running on
   // a system that actually has a service manager.

   WrapScHandle scm(OpenSCManager(NULL, NULL, SC_MANAGER_ALL_ACCESS));
   if (scm == NULL) {
      throw new LastErrorException("Unable to open service manager");
   }
   WrapScHandle service(OpenService(scm, "NTPTime", SERVICE_STOP));
   if (service == NULL) {
      throw new LastErrorException("Unable to open NTPTime service");
   }
   SERVICE_STATUS ss;
   if (! ControlService(service, SERVICE_CONTROL_STOP, &ss)) {
      throw new LastErrorException("Unable to stop NTPTime service");
   }
   return;
}

// ******************************************************* StartNTPTimeService

static void
StartNTPTimeService() {

   // Presumably something in here will fail if we aren't running on
   // a system that actually has a service manager.

   WrapScHandle scm(OpenSCManager(NULL, NULL, SC_MANAGER_ALL_ACCESS));
   if (scm == NULL) {
      throw new LastErrorException("Unable to open service manager");
   }
   WrapScHandle service(OpenService(scm, "NTPTime", SERVICE_START));
   if (service == NULL) {
      throw new LastErrorException("Unable to open NTPTime service");
   }
   if (! StartService(service, 0, NULL)) {
      throw new LastErrorException("Unable to start NTPTime service");
   }
   return;
}

// ********************************************************* StopNTPTimeWindow

static BOOL CALLBACK
StopNTPTimeWindowCallback(HWND hwnd, LPARAM lParam) {
   char buf[100];
   if (GetWindowText(hwnd, buf, sizeof(buf)) > 0) {
      if (strcmp(buf, "NTPTime") == 0) {
         PostMessage(hwnd, WM_CLOSE, 0, 0);
         *((bool *)lParam) = true;
         return FALSE;
      }
   }
   return TRUE;
}

static void
StopNTPTimeWindow() {
   bool closed_it = false;
   EnumWindows(StopNTPTimeWindowCallback, (LPARAM)&closed_it);
   if (! closed_it) {
      throw new SimpleException("NTPTime is already stopped");
   }
   return;
}

// ************************************************** StartNTPTimeControlPanel

static void
StartNTPTimeControlPanel() {
   RunOtherProgram("control.exe", "ntptime.cpl", false);
}

// ******************************************************** StartNTPTimeWindow

static void
StartNTPTimeWindow() {
   RunOtherProgram("ntptime.exe", NULL, false);
}

// *********************************************************** ThisIsWindowsNT

static bool
ThisIsWindowsNT() {
   OSVERSIONINFO osvi;
   osvi.dwOSVersionInfoSize = sizeof(OSVERSIONINFO);
   return GetVersionEx(&osvi) && (osvi.dwPlatformId == VER_PLATFORM_WIN32_NT);
}

// ************************************************************* BrowseNtpLogs

static void
BrowseNtpLogs() {

   // First find out if we even have logging enabled.

   HKEY k;
   DWORD err;
   err = RegOpenKeyEx(HKEY_LOCAL_MACHINE, "SOFTWARE\\ntcli\\NTPTime\\4.3",
                      0, KEY_READ, &k);
   if (err != ERROR_SUCCESS) {
      throw new LastErrorException(err, "Unable to read NTPTime registry key");
   }
   WrapRegKey wrk(k);
   DWORD write_log_flag;
   DWORD len = sizeof(write_log_flag);
   DWORD fmt;
   if ((RegQueryValueEx(k, "WriteLogEntries", 0, &fmt,
                        (LPBYTE)&write_log_flag, &len) != ERROR_SUCCESS) ||
       (fmt != REG_DWORD) ||
       (write_log_flag == 0)) {
      throw new
         SimpleException("WriteLogEntries not enabled in NTPTime registry");
   }

   // If we made it this far, we have logging enabled, so find the
   // log directory name.

   char * logdir = NULL;
   if ((RegQueryValueEx(k, "LogDirectory", 0, &fmt, NULL, &len) !=
        ERROR_SUCCESS) ||
       ((fmt != REG_SZ) && (fmt != REG_EXPAND_SZ))) {
      throw new
         SimpleException("unable to read LogDirectory in NTPTime registry");
   }
   logdir = new char [len + 1];
   if ((RegQueryValueEx(k, "LogDirectory", 0, &fmt, (LPBYTE)logdir, &len) !=
        ERROR_SUCCESS) ||
       ((fmt != REG_SZ) && (fmt != REG_EXPAND_SZ))) {
      delete [] logdir;
      throw new
         SimpleException("unable to read LogDirectory in NTPTime registry");
   }
   if (fmt == REG_EXPAND_SZ) {
      char dummy[1];
      char * origlogdir = logdir;
      DWORD newlen = ExpandEnvironmentStrings(logdir, dummy, 1);
      if (newlen > 0) {
         char * newlogdir = new char [newlen + 1];
         DWORD finlen = ExpandEnvironmentStrings(logdir, newlogdir, newlen);
         delete [] logdir;
         if (newlen >= finlen) {
            logdir = newlogdir;
         } else {
            delete [] newlogdir;
            throw new
               SimpleException("Unable to expand environment in LogDirectory");
         }
      } else {
         delete [] logdir;
         throw new
            SimpleException("Unable to expand environment in LogDirectory");
      }
   }

   char logfile[1024];
   logfile[0] = '\0';

   // PUT CODE HERE - find the most recently written log file and
   // make it the default selection.

   OPENFILENAME ofn;
   ZeroMemory(&ofn, sizeof(ofn));
   ofn.lStructSize = sizeof(ofn);
   ofn.hwndOwner = main_window;
   ofn.lpstrFilter = "NTPTime Log Files\0*.ntp\0All Files\0*.*\0\0";
   ofn.lpstrInitialDir = logdir;
   ofn.lpstrTitle = "Open NTPTime Log File";
   ofn.Flags = OFN_FILEMUSTEXIST|OFN_NOCHANGEDIR|OFN_PATHMUSTEXIST;
   ofn.lpstrFile = logfile;
   ofn.nMaxFile = sizeof(logfile);
   if (GetOpenFileName(&ofn)) {
      RunOtherProgram(log_viewer, logfile, true);
   }
   return;
}

// *********************************************************** RunOtherProgram

static void
RunOtherProgram(char * progname, char * opt_arg, bool quote_arg) {
   char fullname[1024];
   char cmdline[256];
   char * lastpart = NULL;
   DWORD arglen = (opt_arg == NULL ? 0 : strlen(opt_arg));
   if (arglen > 0) {
      arglen++;                         // separator
      if (quote_arg) arglen += 2;       // quotes
      arglen++;                         // null byte
   }
   DWORD plen = SearchPath(NULL, progname, NULL, sizeof(fullname),
                           fullname, &lastpart);
   if (plen == 0) {
      throw new LastErrorException("SearchPath unable to locate \"%s\"",
                                   progname);
   }
   if ((plen == 0) || (plen >= sizeof(fullname)) || (lastpart == NULL) ||
       ((strlen(lastpart) + arglen) >= sizeof(cmdline))) {
      throw new SimpleException("full path to \"%s\" too long", progname);
   }
   strcpy(cmdline, lastpart);
   if (opt_arg != NULL) {
      strcat(cmdline, " ");
      if (quote_arg) {
         strcat(cmdline, "\"");
      }
      strcat(cmdline, opt_arg);
      if (quote_arg) {
         strcat(cmdline, "\"");
      }
   }
   STARTUPINFO si;
   ZeroMemory((void *)&si, sizeof(si));
   si.cb = sizeof(si);
   PROCESS_INFORMATION pi;
   if (! CreateProcess(fullname, cmdline, NULL, NULL, FALSE,
                       CREATE_DEFAULT_ERROR_MODE|
                          CREATE_NEW_PROCESS_GROUP|
                             NORMAL_PRIORITY_CLASS,
                       NULL, NULL, &si, &pi)) {
      throw new LastErrorException("CreateProcess failed for %s", cmdline);
   }

   // Avoid leaking process and thread handles...

   CloseHandle(pi.hProcess);
   CloseHandle(pi.hThread);
   return;
}
